/*
  smtppipe 
  see also: http://www.ogris.de/smtppipe/

  smtppipe --help gives you some help
  or look at usage() below

  this software is freely distributable
  as described by the terms of the GNU GPL;
  see http://www.gnu.org/licenses/gpl.html
*/

#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <resolv.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <pwd.h>

#define VERSION "0.10"

#define STDIN 0
#define STDOUT 1
#define STDERR 2

#define TCP "tcp"
#define SMTP "smtp"
#define BLKSIZE 4096

int sock = -1;
char *mail;

void die (const char *s1, const char *s2)
{
  if (mail) free(mail);
  if (sock >= 0) close(sock);

  write(STDERR, s1, strlen(s1));
  if (s2) {
    write(STDERR, ": ", 2);
    write(STDERR, s2, strlen(s2));
  }
  write(STDERR, "\n", 1);
  exit(1);
}

char *strnchr (char *haystack, char needle, size_t n)
{
  static size_t i;

  for (i = 0; (i < n) && *(haystack + i); ++i)
    if (*(haystack + i) == needle) return haystack + i;
  return NULL;
}

char *strnstr (char *haystack, char *needle, size_t n)
{
  static size_t i, j, len;

  len = strlen(needle);

  for (i = 0; (i <= n - len) && *(haystack + i); ++i) {
    for (j = 0; j < len; ++j) if (*(haystack + i + j) != *(needle + j)) break;
    if (j == len) return haystack + i;
  }

  return NULL;
}
void usage ()
{
#define HELP \
"smtppipe "VERSION" - reads mail from stdin and forwards it to a smtp relay\n" \
"conformal with sendmail and qmail-inject command line switches\n\n" \
"smtppipe [-a] [-h] [-t] [-A] [-H] [-f sender] [-m smtprelay] [recipient ... ]\n" \
"smtppipe --help\n\n" \
"  -f         forces envelope sender address;\n" \
"             default: mail header from address; last ressort: current user name\n" \
"  -m         host name or address of a smtp relay;\n" \
"             default: value of the environment variable SMTPRELAY\n" \
"  -a         send message to all addresses given as recipients arguments;\n" \
"             do not use any header recipient addresses\n" \
"  -h, -t     send message to all header recipient addresses (To, Cc, Bcc);\n" \
"             do not use any command line recipients\n" \
"  -A         send message to all addresses given as recipient arguments; if\n" \
"             no recipient arguments are given, use header recipient addresses;\n" \
"             this is the default behaviour\n" \
"  -H         send message to all addresses given as recipient arguments and\n" \
"             to all addresses given as header recipients\n" \
"  recipient  list of space seperated recipient addresses\n" \
"  --help     displays this help\n\n" \
"see http://www.ogris.de/smtppipe/\n"

  write(STDOUT, HELP, strlen(HELP));
  exit(1);
}

char *stralloc (char *p, size_t len)
{
  p = realloc(p, len);
  if (!p) die("realloc()", strerror(errno));
  return p;
}

size_t stralloc_cpy (char **dst, char *src)
{
  static size_t len;

  len = strlen(src);
  *dst = stralloc(NULL, len + 1);
  strcpy(*dst, src);
  return len;
}

size_t stralloc_cat (char **dst, char *src)
{
  static size_t len;

  len = strlen(*dst) + strlen(src);
  *dst = stralloc(*dst, len + 1);
  strcat(*dst, src);
  return len;
}

void w2s (int socket, char *command, char *value, char *code)
{
  static ssize_t ret;
  static size_t len;
  static char *buf = NULL;

  if (command) {
    stralloc_cpy(&buf, command);

    if (value) {
      stralloc_cat(&buf, " ");
      stralloc_cat(&buf, value);
    }

    len = stralloc_cat(&buf, "\r\n");

    ret = write(socket, buf, len);
    if (ret < 0) die("write(socket)", strerror(errno));
    if ((size_t) ret != len)
      die("write(socket)", "could not write whole data stream");
  }

  if (code) {
    buf = stralloc(buf, BLKSIZE);

    ret = read(socket, buf, BLKSIZE);
    if (ret < 0) die("read(socket)", strerror(errno));

    if (strncmp(buf, code, strlen(code))) {

      /* trim the return string */
      for (; (*(buf + ret - 1) == '\r') || (*(buf + ret - 1) == '\n'); --ret)
        *(buf + ret - 1) = '\0';

      write(STDERR, "The smtp relay returns an unexpected status code:\n", 50);
      write(STDERR, buf, ret);
      if (command) {
        write(STDERR, "\nThe requested command was:\n", 28);
        write(STDERR, command, strlen(command));
        if (value) {
          write(STDERR, " ", 1);
          write(STDERR, value, strlen(value));
        }
      }
      if (code) {
        write(STDERR, "\nThe expected status code was: ", 31);
        write(STDERR, code, strlen(code));
      }
      free(buf);
      die("\n", NULL);
    }
  }
  if (buf) free(buf);
}

void pw2sock (int sock, char **from, char *hostname)
{
  struct passwd *pw;

  pw = getpwuid(getuid());
  if (!pw) {
    w2s(sock, "QUIT", NULL, NULL);
    die("can not figure out your sender's address", NULL);
  }
  stralloc_cpy(from, pw->pw_name);
  stralloc_cat(from, "@");
  stralloc_cat(from, hostname);
  w2s(sock, "MAIL FROM:", *from, "250");
  free(*from);
}

int main (int argc, char *argv[])
{
  struct protoent *pe;
  struct servent *se;
  struct hostent *he;
  struct sockaddr_in sin;
  size_t maillen, cnt_nl;
  int i, recips, mode, sock;
  char *p, *sep, *left, *right, *bccleft, *from, *relay;

  for (i = 1, recips = 0, mode = 0, from = NULL, relay = NULL; i < argc; i++) {
    if ( (!strncmp(argv[i], "--h", 3)) || (!strncmp(argv[i], "-he", 3)) )
      usage();

    if (!strcmp(argv[i], "--")) continue;

    if (!strncmp(argv[i], "-f", 2)) {
      if (strlen(argv[i]) > 2) {
        from = argv[i]+2;
        continue;
      }
      if (i < argc - 1) {
        from = argv[++i];
        continue;
      }
      die("-f requires an envelope from-address", NULL);
    }

    if (!strncmp(argv[i], "-m", 2)) {
      if (strlen(argv[i]) > 2) {
        relay = argv[i]+2;
        continue;
      }
      if (i < argc - 1) {
        relay = argv[++i];
        continue;
      }
      die("-m requires a host name or address", NULL);
    }

    if (!strcmp(argv[i], "-a")) {
      mode = 1;
      continue;
    }

    if ( (!strcmp(argv[i], "-h")) || (!strcmp(argv[i], "-t")) ) {
      mode = 3;
      continue;
    }

    if (!strcmp(argv[i], "-A")) {
      mode = 0;
      continue;
    }

    if (!strcmp(argv[i], "-H")) {
      mode = 2;
      continue;
    }

    if (strncmp(argv[i], "-", 1)) {
      recips = i;
      break;
    }

    die("unknown parameter", argv[i]);
  }

  /* must have an smtprelay */
  if ((!relay) && (!(relay = getenv("SMTPRELAY"))))
    die("no smtprelay specified", NULL);

  if (!(he = gethostbyname(relay))) {
    write(STDERR, "gethostbyname(", 14);
    write(STDERR, relay, strlen(relay));
    die(")", hstrerror(h_errno));
  }

  if (!(pe = getprotobyname(TCP)))
    die("unknown protocol", TCP);

  if (!(se = getservbyname(SMTP, pe->p_name)))
    die("unknown service", SMTP);

  memcpy((char*) &sin.sin_addr, he->h_addr, he->h_length);
  sin.sin_port = se->s_port;
  sin.sin_family = PF_INET;

  /*
   * keep our hostname for the initial SMTP HELO below and
   * to form the sender's address if not specified by commandline parameter or
   * by header from address
   */
  mail = stralloc(NULL, BLKSIZE);

  if (gethostname(mail, BLKSIZE))
    die("gethostname()", strerror(errno));

  if (!(he = gethostbyname(mail))) {
    write(STDERR, "gethostbyname(", 14);
    write(STDERR, mail, strlen(mail));
    die(")", hstrerror(h_errno));
  }

  /* get the mail through STDIN */
  /* insert an '\r' before each '\n' */
  for (maillen = 0;;) {
    mail = stralloc(mail, maillen + BLKSIZE);

    i = read(STDIN, mail + maillen, BLKSIZE);
    if (!i) break;
    if (i < 0) die("read(STDIN)", strerror(errno));

    /* count all '\n' w/o prepending '\r' */
    for (p = mail + maillen, cnt_nl = 0; p < mail + maillen + i; ++p) {
      if (*p != '\n') continue;
      if (p == mail) die("empty line at the beginning of your mail", NULL);
      if (*(p - 1) == '\r') continue;
      ++cnt_nl;
    }

    mail = stralloc(mail, maillen + i + cnt_nl);

    /* insert a '\r' before those '\n' */
    for (p = mail + maillen; p < mail + maillen + i; ++p) {
      if (*p != '\n') continue;
      if (*(p - 1) == '\r') continue;
      memmove(p + 1, p, i - (p - mail - maillen));
      ++maillen;
      *p++ = '\r';
    }

    maillen += i;
  }

  /* no incomplete last line */
  if (strcmp(mail + maillen - 2, "\r\n")) {
    mail = stralloc(mail, maillen + 2);
    *(mail + maillen) = '\r';
    *(mail + maillen + 1) = '\n';
    maillen += 2;
  }

  /* escape single dots on a line */
  for (p = strnstr(mail, "\r\n.\r\n", maillen); p;
       p = strnstr(p, "\r\n.\r\n", maillen - (p - mail))) {
    i = p - mail;
    mail = stralloc(mail, ++maillen);
    p = mail + i;  // mail may have been reallocated
    p += 3;
    memmove(p + 1, p, maillen - 1 - (p - mail));
    *p = '.';
    ++p;
  }

  /* no header, no body? */
  sep = strnstr(mail, "\r\n\r\n", maillen);
  if (!sep) die("no body seperation found", NULL);

  /* connect to server */
  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    die("socket()", strerror(errno));

  if (connect(sock, (struct sockaddr*) &sin, sizeof(sin)))
    die("connect()", strerror(errno));

  /* smtp blabla */
  w2s(sock, NULL, NULL, "220");
  w2s(sock, "HELO", he->h_name, "250");
  if (from) w2s(sock, "MAIL FROM:", from, "250");

  /* look for header from address, evaluate header recips, erase bcc's */
  for (p = mail; p < sep;) {

    if (!strncasecmp(p, "from: ", 6)) {
      if (!from) {
        from = strchr(p, ' ');

        /* !p cant happen because we set sep to boby seperation above */
        p = strchr(++from, '\r');
        *p = '\0';
        w2s(sock, "MAIL FROM:", from, "250");
        *p++ = '\r';
      }
      else p = strchr(p, '\n');
    }

    else if ( (!strncasecmp(p, "to: ", 4)) ||
              (!strncasecmp(p, "cc: ", 4)) ||
              (!strncasecmp(p, "bcc: ", 5)) ) {
      bccleft = p;
      left = strchr(p, ' ');

      /* !p cant happen because we set sep to body seperation above */
      p = strchr(++left, '\r');

      if ((mode > 1) || (!mode && (recips <= 0))) {
        for (; left < p;) {

          /* last ressort: guess sender's address from current uid */
          if (!from) pw2sock(sock, &from, he->h_name); //{

          /* now we can savely print out our recipients wish list */
          right = strnstr(left, ", ", maillen - (mail - left));
          if (!right || (right > p)) right = p;
          *right = '\0';
          w2s(sock, "RCPT TO:", left, "250");
          *right = ',';
          left = right + 2;
          if (!recips) recips = -1;
        }
      }

      /* maybe right was set to p and *p was set to ',' */
      *p++ = '\r';

      /* erase bcc line anyway, avoid memory trashing */
      if (!strncasecmp(bccleft, "bcc: ", 5)) {
        memmove(bccleft, p + 1, maillen - (p + 1 - mail));
        maillen -= p + 1 - bccleft;
        sep -= p + 1 - bccleft;
        mail = stralloc(mail, maillen);
      }
    }

    else p = strchr(p, '\n');

    ++p;
  } /* for */

  /* evaluate cmd line recipients */
  if ((mode < 3) && (recips > 0)) {
    if (!from) pw2sock(sock, &from, he->h_name);
    for (i = recips; i < argc; i++) w2s(sock, "RCPT TO:", argv[i], "250");
  }

  /* no recipients at all? */
  if (!recips) {
    w2s(sock, "QUIT", NULL, NULL);
    die("no recipients found at all", NULL);
  }

  /* send the beast */
  w2s(sock, "DATA", NULL, "354");

  i = write(sock, mail, maillen);
  if (i < 0) die("write(socket)", strerror(errno));
  if ((size_t) i != maillen)
    die("write(socket)", "could not write the whole data stream");

  w2s(sock, ".", NULL, "250");
  w2s(sock, "QUIT", NULL, NULL);
  close(sock);

  free(mail);

  return(0);
}
